// convolution-test.cc

#include "splines/Grid.h"
#include "splines/Spline.h"

#include <iostream>
#include <string>
#include <cmath>

using namespace splines;

#define VERBOSE_checkConvolution
//#define TRACE_checkConvolution
//#define BRIEF_checkConvolution

//#define VERBOSE_checkSpline

template <class S, class F> 
bool
checkSpline ( S const & s, F const & f, 
		double start, double end, double h, std::string message ) 
{
  bool ok = true;
  std::vector<double> vv(1);  // made this 2 to test exception behavior
  double xx = (3*start+end)/4;
  vv[0] = xx; 
  if ( s(xx) != s(vv) ) {
    std::cout << message << ":\n    s(xx) = " << s(xx) << " s(vv) =  "
	      << s(vv) << "\n";
    ok = false;
  }
  double h4 = h*h*h*h;
  double worst = 0;
  double x0;
  double x = start;
//  const int intervals = 4357;
  const int intervals = 43;
  double spacing = (end-start)/intervals;
  while (x < end) {
    double v_s = s(x);
    double v_f = f(x);
    double err = fabs(v_s - v_f);
    double scaled_err = err / h4; 
    #ifdef VERBOSE_checkSpline
    std::cout << message << ":\n    x = " << x 
              << " s = " << v_s << " f = " << v_f 
              << " error = " << err << "\n";
    #endif
    if ( err > worst ) {
      x0 = x;
      worst = err;
      			#ifdef TRACE_checkSpline
			std::cout << message << "  " << x << "  "
				  << scaled_err << "\n";
			#endif
      if ( scaled_err > 20 ) {
	#ifndef VERBOSE_checkSpline
	#ifndef TRACE_checkSpline
        if ( err > 1.0e-3 )  {
          std::cout << message << ":\n    x = " << x << " error = "
		  << err << " which scales to " << scaled_err << "\n";
	}
        #endif
	#endif
	ok = false;
      }
    }
    x += spacing;
  }
  #ifdef BRIEF_checkSpline
  std::cout << message << ": worst = spline error = " << worst <<"\n";
  #endif
  return ok;
}


static bool testOneConvolution ( Spline<1> const & f,
                                 Spline<1> const & g, 
                                 double fmin, double fmax,
                                 Spline<1> const & expectation ) 
{
  Spline<1> s(f.convolution(g));
  for (int i=0; i<101; ++i) {
    double x = fmin + .01*i*(fmax-fmin);
    double value = s(x);
    double expected_value = expectation(x); 
    if ( std::abs (value - expected_value) > .0001 ) 
    {
      std::cout << "Unexpected value for convolution at x = " << x << ": \n"
                << "s(" << x << ") = " << value 
                << " (expected " << expected_value << ")\n";
      // return false;
    }
  }  
  return true;
}

static bool testOneConvolution ( Spline<1> const & f,
                                 Function_1 const & g, 
                                 double fmin, double fmax,
                                 double gmin, double gmax,
                                 Spline<1> const & expectation ) 
{
  Spline<1> s(f.convolution(g,gmin,gmax));
  for (int i=0; i<101; ++i) {
    double x = fmin + .01*i*(fmax-fmin);
    double value = s(x);
    double expected_value = expectation(x); 
    if ( std::abs (value - expected_value) > .0001 ) 
    {
      std::cout << "Unexpected value for convolution at x = " << x << ": \n"
                << "s(" << x << ") = " << value 
                << " (expected " << expected_value << ")\n";
      // return false;
    }
  }  
  return true;
}

const double h = 2.0;

double sin2(double x) {
  double a = std::sin(x);
  return a*a;
}

double triangleFunction (double x) {
  const double a = -1.0/h;
  const double b = +1.0/h;
  const double s = h/b;
  if ( x < a ) return 0;
  if ( x < 0 ) return s*(x-a);
  if ( x < b ) return h - s*x;
  return 0;
}

double convolutedSin2 (double x) {
  double sh = std::sin(1.0/h);
  return 0.5*(1 - h*h*sh*sh*std::cos(2*x));
}

bool conv () 
{ 
  bool ok = true;
  double splineGridSpacing = .05;
  Grid<1>   splineGrid  ( -13.0, 13.0 + splineGridSpacing, 
        1 + 26/splineGridSpacing + .5 );
  Spline<1> sin2Spline  (sin2, splineGrid);
  double smearGridSpacing = .005;
  Grid<1>   smearGrid   ( -2/h, 2/h + smearGridSpacing, 
        1 + 4/(h*smearGridSpacing) + .5 );
  Spline<1> smearSpline ( triangleFunction, smearGrid );
//#define SPECIAL_GIBBS_CHECK
#ifdef  SPECIAL_GIBBS_CHECK  
  { double x = -1/h +  .5 * smearGridSpacing;
    double difference = smearSpline(x) - triangleFunction(x);
    std::cout << "At x = " << x
              << " smearSpline =  " << smearSpline(x)
              << " triangleFucntion = " << triangleFunction(x)
              << " difference = " << difference
              << "\n"
              << " scales to " << difference/smearGridSpacing << " "
              << difference/(smearGridSpacing*smearGridSpacing) << " "
              << difference/(smearGridSpacing*smearGridSpacing*smearGridSpacing) 
              << " " << difference/
              (smearGridSpacing*smearGridSpacing*
              smearGridSpacing*smearGridSpacing) << "\n"; 
  }
#endif
  checkSpline( sin2Spline, sin2, -8.0, 8.0, splineGridSpacing, "sin^2 spline" );
  checkSpline( smearSpline, triangleFunction, -1.0, 1.0, smearGridSpacing, 
                                                "triangleFunction" );
  Spline<1> expectation ( convolutedSin2, splineGrid );
  ok &= testOneConvolution ( sin2Spline, smearSpline, -8, 8, expectation );
  ok &= testOneConvolution ( sin2Spline, triangleFunction, -8, 8, 
                             -1,1, expectation );
  return ok;
}

int main() {
  bool ok = true;
  ok &= conv();
  return ok ? 0:-1;
}
