// grid-1.cpp

#include "splines/Grid.h"
#include "splines/GridSearchAnswer.h"

#include <iostream>
#include <string>
#include <cmath>

using namespace splines;

bool checkAnswer ( GridSearchAnswer & a, bool res,
		   int index, double x, double h, bool r,
		   std::string const & msg,
		   double value ) {
   bool ok = true;
  if (a.index != index) { 
    std::cout << msg << value << " index should be " << index 
              << "; is " << a.index << "\n";
    ok =  false;
  }
  if (fabs (a.x - x) > 1.0E-12) {
    std::cout << msg << value << " x should be " << x 
              << "; is " << a.x << "\n";
    ok =  false;
  }
  if (fabs (a.h - h) > 1.0E-12) {
    std::cout << msg << value << " h should be " << h 
              << "; is " << a.h << "\n";
    ok =  false;
  }
  if ( res != r ) {
    std::cout << msg << value << " return value should be " << r 
              << "; is " << res << "\n";
    ok =  false;
  }
  return ok;		   
}

bool oneUniformRegion ( double start, double end, int n ) {
  Grid<1> g(start, end, n);
  double spacing = (end-start)/n;
  double p;
  bool  in;
  bool retval = true;
  GridSearchAnswer a;
  // to the left of start should return false
  p  = start - .75 * spacing;
  in = g.search ( p, a );
  retval &= checkAnswer ( a, in, 0, -.75, spacing,  false, 
  					"1UR p  = start - ", .75*spacing );
  // right at each grid point (up to penultimate) should give that index
  // 1/4 way thru intervals (up to penultimate) should give x=.25
  for (int i=0; i < n-1; ++i) {
    p = start + i * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, 0, spacing, true, 
    					"1UR p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, offset, spacing, true, 
    					"1UR p = ", p ); 
  } 
  // At last actual point
  p = end - spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 1.0, spacing, true, 
    					"1UR p = end - spacing = ", p ); 
  // between last actual point and end
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 1.625, spacing, false, 
    				     "1UR p = end - spacing + ", .625*spacing ); 
  // at end 
  p = end;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 2.0, spacing, false, 
    					"1UR p = end = ", p ); 
  // Past end 
  p = end + .5 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 2.5, spacing, false, 
    					"1UR p = end + ", .5*spacing ); 
  return retval;  
}

bool oneIrregularRegion ( std::vector<double> pts ) {
  Grid<1> g(pts.begin(), pts.end());
  double spacing;
  double p;
  bool  in;
  int n = pts.size();
  bool retval = true;
  GridSearchAnswer a;
  // to the left of start should return false
  spacing = pts[1] - pts[0];
  p  = pts[0] - .75 * spacing;
  in = g.search ( p, a );
  retval &= checkAnswer ( a, in, 0, -.75, spacing,  false, 
  					"1IR p  = start - ", .75*spacing );
  // right at each grid point (up to penultimate) should give that index
  // 1/4 way thru intervals (up to penultimate) should give x=.25
  for (int i=0; i < n-1; ++i) {
    spacing = pts[i+1] - pts[i];
    p = pts[i];
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, 0, spacing, true, 
    					"1IR p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, offset, spacing, true, 
    					"1IR p = ", p ); 
  } 
  // At last actual point
  spacing = pts[n-1] - pts[n-2];
  p = pts[n-1];
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 1.0, spacing, true, 
    					"1UR p = end - spacing = ", p ); 
  // past last actual point and end
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2, 1.625, spacing, false, 
    				     "1UR p = end - spacing + ", .625*spacing ); 
  return retval;  
}

bool mixedGrid ( std::vector<double> p1, 
		 double start2, double end2, int n2,
		 double start3, double end3, int n3,
		 std::vector<double> p4, std::vector<double> p5 ) {
  Grid<1> g1(p1.begin(), p1.end());
  Grid<1> g2(start2, end2, n2);
  Grid<1> g3(start3, end3, n3);
  Grid<1> g4(p4.begin(), p4.end());
  Grid<1> g5(p5.begin(), p5.end());
  Grid<1> g = g1+g2+g3+g4+g5;
  double spacing;
  double p;
  bool  in;
  int n;
  int i0 = 0;
  bool retval = true;
  GridSearchAnswer a;
  // --- Before g1 ---
  n = p1.size();
  // to the left of start should return false
  spacing = p1[1] - p1[0];
  p  = p1[0] - .75 * spacing;
  in = g.search ( p, a );
  retval &= checkAnswer ( a, in, 0, -.75, spacing,  false, 
  					"mG1 p  = start - ", .75*spacing );
  // --- In g1 ---
  for (int i=0; i < n-1; ++i) {
    spacing = p1[i+1] - p1[i];
    p = p1[i];
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, 0, spacing, true, 
    					"mG1 p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i, offset, spacing, true, 
    					"mG1 p = ", p ); 
  } 
  // --- Between g1 and g2 ---
  // At last actual point of g1
  spacing = start2 - p1[n-1];
  p = p1[n-1];
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1, 0, spacing, true, 
    					"mG12 p = p1[n-1] = ", p ); 
  // past last actual point of g1
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1, 0.625, spacing, true, 
    				     "mG12 p = p1[n-1]  + ", .625*spacing ); 
  // --- In g2 ---
  i0 += n; 
  n = n2;
  spacing = (end2-start2)/n;
  for (int i=0; i < n-1; ++i) {
    p = start2 + i * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, 0, spacing, true, 
    					"mG2 p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, offset, spacing, true, 
    					"mG2 p = ", p ); 
  } 
  // --- Between g2 and g3 ---
  // At last actual point of g2
  p = end2 - spacing;
  spacing = start3 - p;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0.0, spacing, true, 
    					"mG23 p = end - spacing = ", p ); 
  // between last actual point and end
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0.625, spacing, true, 
    				     "mG23 p = end - spacing + ", .625*spacing ); 
  // --- In g3 ---
  i0 += n; 
  n = n3;
  spacing = (end3-start3)/n;
  for (int i=0; i < n-1; ++i) {
    p = start3 + i * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, 0, spacing, true, 
    					"mG3 p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, offset, spacing, true, 
    					"mG3 p = ", p ); 
  } 
  // --- Between g3 and g4 ---
  // At last actual point of g3
  p = end3 - spacing;
  spacing = p4[0] - p;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0.0, spacing, true, 
    					"mG34 p = end - spacing = ", p ); 
  // between last actual point and end
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0.625, spacing, true, 
    				     "mG34 p = end - spacing + ", .625*spacing ); 
  // --- In g4 ---
  i0 += n; 
  n = p4.size();
  for (int i=0; i < n-1; ++i) {
    spacing = p4[i+1] - p4[i];
    p = p4[i];
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, 0, spacing, true, 
    					"mG4 p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, offset, spacing, true, 
    					"mG4 p = ", p ); 
  } 
  // --- Between g4 and g5 ---
  // At last actual point of g4
  spacing = p5[0] - p4[n-1];
  p = p4[n-1];
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0, spacing, true, 
    					"mG45 p = p4[n-1] = ", p ); 
  // past last actual point of g5
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-1+i0, 0.625, spacing, true, 
    				     "mG45 p = p4[n-1]  + ", .625*spacing ); 
  // --- In g5 ---
  i0 += n;
  n = p5.size();
  for (int i=0; i < n-1; ++i) {
    spacing = p5[i+1] - p5[i];
    p = p5[i];
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, 0, spacing, true, 
    					"mG5 p = exactly ", p ); 
    double offset = .25;
    p += offset * spacing;
    in = g.search ( p, a ); 
    retval &= checkAnswer ( a, in, i+i0, offset, spacing, true, 
    					"mG5 p = ", p ); 
  } 
  // --- After g5 ---
  // At last actual point
  spacing = p5[n-1] - p5[n-2];
  p = p5[n-1];
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2+i0, 1.0, spacing, true, 
    					"mg5+ p = end - spacing = ", p ); 
  // past last actual point and end
  p += .625 * spacing;
  in = g.search ( p, a ); 
  retval &= checkAnswer ( a, in, n-2+i0, 1.625, spacing, false, 
    				  "mG5+ p = end - spacing + ", .625*spacing ); 
  return retval;  
}

bool shift ( std::vector<double> p1, 
		 double start2, double end2, int n2,
		 double start3, double end3, int n3,
		 std::vector<double> p4, std::vector<double> p5,
                 double delta ) {
  Grid<1> g1(p1.begin(), p1.end());
  Grid<1> g2(start2, end2, n2);
  Grid<1> g3(start3, end3, n3);
  Grid<1> g4(p4.begin(), p4.end());
  Grid<1> g5(p5.begin(), p5.end());
  Grid<1> g = g1+g2+g3+g4+g5;
  Grid<1> h(g);
  h.shift(delta);
  if (h.nPoints() != g.nPoints()) {
    std::cout << "nPoints discrepancy in shift()\n";
    return false;
  }
  for (unsigned int i=0; i < g.nPoints(); ++i) {
    #ifdef OUTPUT_BOTH_GRIDS
    std::cout << "g[" << i << "] = " << g[i]
              << " h[" << i << "] = " << h[i] << "\n";
    #endif
    if (std::abs( h[i] - (g[i]+delta) ) > 1.0e-12) {
      std::cout << " Value discrepancy in shift():\n"
                << "h[" << i << "] = " << h[i]
                << " g[" << i << "] + " << delta << " = " << g[i]+delta
                << " (difference = " << h[i] - (g[i]+delta) << ")\n";
      return false;
    } 
  } 
  return true;
}  
int main() {
  if (!oneUniformRegion ( -100.0, 50.0, 50 )) return -1;
  std::vector<double> pts;
  for (int i=0; i<50; ++i) {
    double x = i*i;
    pts.push_back(x);
  }
  if (!oneIrregularRegion (pts)) return -1;
  std::vector<double> p1;
  std::vector<double> p4;
  std::vector<double> p5;
  for (int i=0; i<20; ++i) {
    double x = i*i;
    p1.push_back(x);
  }
  double start2 = 460; double end2 = 640; int n2 = 24;
  double start3 = 638; double end3 = 938; int n3 = 15;
  for (int i=0; i<10; ++i) {
    double x = 950 + i*i;
    p4.push_back(x);
  }
  for (int i=0; i<12; ++i) {
    double x = 1110 + i*i;
    p5.push_back(x);
  }
  if (!mixedGrid (p1, start2, end2, n2, start3, end3, n3, p4, p5)) return -1;
  if (!shift(p1, start2, end2, n2, start3, end3, n3, p4, p5, 9.0)) return -1;

//  std::cout << "All Grid-1 tests passed\n";
  return 0;
}
