// integrate-1-test.cc

#include "splines/Grid.h"
#include "splines/Spline.h"

#include <iostream>
#include <string>
#include <cmath>

using namespace splines;

//#define VERBOSE_checkIntegrate
//#define TRACE_checkIntegrate
//#define BRIEF_checkIntegrate


static bool testOneDefiniteIntegralValue ( Spline<1> const & s, 
           double a, double b, double expectation, std::string const & id ) 
{
  double value = s.integrate(a,b);
  if ( std::abs (value - expectation) > .000001 ) 
  {
    #ifdef TRACE_checkIntegrate
    std::cout << "integral of "  << id <<  "\n"
              << "s.integral(" << a << ", " << b << ") = " << value 
              << " (delta = " << expectation - value << ")\n";
    
    #endif
    std::cout << "Unexpected value for integral: "  << id <<  "\n"
              << "s.integral(" << a << ", " << b << ") = " << value 
              << " (expected " << expectation << ")\n";
    return false;
  }  
  return true;
}

template <class S, class F> 
bool
checkSpline ( S const & s, F const & f, 
		double start, double end, double h, std::string message ) 
{
  bool ok = true;
  std::vector<double> vv(1);  // made this 2 to test exception behavior
  double xx = (3*start+end)/4;
  vv[0] = xx; 
  if ( s(xx) != s(vv) ) {
    std::cout << message << ":\n    s(xx) = " << s(xx) << " s(vv) =  "
	      << s(vv) << "\n";
    ok = false;
  }
  double h4 = h*h*h*h;
  double worst = 0;
  double x0;
  double x = start;
  const int intervals = 4357;
  double spacing = (end-start)/intervals;
  #ifdef VERBOSE_checkIntegrate
    std::cout << message << "\n";
    int oldPrecision = std::cout.precision(12);
  #endif
  while (x < end) {
    double v_s = s(x);
    double v_f = f(x);
    double err = fabs(v_s - v_f);
    double scaled_err = err / h4; 
    #ifdef VERBOSE_checkIntegrate
//    std::cout << " x " << x << " s(x) " << v_s 
//              << " f(x) " << v_f << " delta " << v_s - v_f << "\n";
    std::cout << x << " " << v_s << " " << v_f << " " << v_s - v_f << "\n";
    #endif
    if ( err > worst ) {
      x0 = x;
      worst = err;
      #ifdef TRACE_checkIntegrate
	std::cout << message << "  " << x << "  "
		  << scaled_err << "\n";
      #endif
      if ( scaled_err > 20 ) {
	#ifndef VERBOSE_checkIntegrate
	#ifndef TRACE_checkIntegrate
        std::cout << message << ":\n    x = " << x << " error = "
		  << err << " which scales to " << scaled_err << "\n";
	#endif
	#endif
	ok = false;
      }
    }
    x += spacing;
  }
  #ifdef BRIEF_checkIntegrate
  std::cout << message << ": worst =  " << worst/h4 <<"\n";
  #endif
  #ifdef VERBOSE_checkIntegrate
    std::cout.precision(oldPrecision);
  #endif

  return ok;
}

const double indefiniteIntegralLowerLimit = .765;


double sin4x(double x) {
  return std::sin(4*x);
}

double x3(double x) {
  return x*x*x;
}

double cos4xfun (double x) { // intended as integral (lowlim,x) sin(4*x) dx 
  return -0.25*( std::cos(4*x) - std::cos(4*indefiniteIntegralLowerLimit) );
}

double x4fun (double x) { // intended as integral (lowlim,x) x^3 dx 
  return 0.25*( std::pow(x,4) - std::pow(indefiniteIntegralLowerLimit,4) );
}

bool definite_integral () 
{ 
  bool ok = true;
  double h = .01;
  Grid<1> g( -5, 5+h, 10/h+1.5); 
  Spline<1> ssin ( sin4x, g );
  Spline<1> sx3  ( x3, g );
  double a = -1;
  double b = 3;
  double expectationx3 = 0.25*(b*b*b*b - a*a*a*a);
  ok &= testOneDefiniteIntegralValue(sx3, a, b, expectationx3,"x^3");  
  double expectations4x = 0.25*(std::cos(4*a)-std::cos(4*b));
  ok &= testOneDefiniteIntegralValue(ssin, a, b, expectations4x,"sin4x");  
  return ok;
}

bool indefinite_integral() {
  bool ok = true;
  double h = .01;
  Grid<1> g( -5, 5+h, 10/h+1.5); 
  Spline<1> ssin ( sin4x, g );
  Spline<1> int_sin4x ( ssin.indefiniteIntegral(indefiniteIntegralLowerLimit) );
  ok &= checkSpline ( int_sin4x, cos4xfun, -1, 3, h, "int_sin4x" );
  Spline<1> sx3  ( x3, g );
  Spline<1> int_sx3 ( sx3.indefiniteIntegral(indefiniteIntegralLowerLimit) );
  ok &= checkSpline ( int_sx3, x4fun, -1, 3, h, "int_sx3" );
  return ok;
}

bool delta_function_integrals() {
  bool ok = true;
  DeltaFunction<1> d( 0, 1, 24.0 );
  Spline<1> sd(d);
  Spline<1> di (sd.indefiniteIntegral(0));
  double xlow = -.5;
  int n = 72;
  //#define LOOK_AT_DELTA_FUNCTION
  #ifdef  LOOK_AT_DELTA_FUNCTION
  std::cout << "Delta function spline \n";
  for (int i=0; i < n+1; ++i) {
    double x = xlow - i * 2*xlow/n;
    std::cout << x << "  " << sd(x) << "\n";
  } 
  std::cout << "Theta function spline \n";
  for (int i=0; i < n+1; ++i) {
    double x = xlow - i * 2*xlow/n;
    std::cout << x << "  " << di(x) << "\n";
  } 
  #endif
  if ( std::abs(di(-.1)) > .000001 ) {
    std::cout << "integral of delta function wrong at -.1: "  
              <<  di(-.1) << "\n";
    ok = false;
  }
  if ( std::abs(di(+.1) - 1.0) > .000001 ) {
    std::cout << "integral of delta function wrong at +.1: "  
              <<  di(.1) << "\n";
    ok = false;
  }
  if ( std::abs(di(0) - .5) > .000001 ) {
    std::cout << "integral of delta function wrong at 0: "  
              <<  di(0) << "\n";
    ok = false;
  }
  return ok;  
}

int main() {
  bool ok = true;
  ok &= definite_integral();
  ok &= indefinite_integral();
  ok &= delta_function_integrals();
  return ok ? 0:-1;
}
