// spline-1.cpp

#include "splines/Grid.h"
#include "splines/Spline.h"

#include <iostream>
#include <string>
#include <cmath>

using namespace splines;

//#define VERBOSE_checkSpline
//#define TRACE_checkSpline
#define BRIEF_checkSpline

const double pi = 3.14159265358979;

template <class S, class F> 
bool
checkSpline ( S const & s, F const & f, 
		double start, double end, double h, std::string message ) 
{
  bool ok = true;
  std::vector<double> vv(1);  // made this 2 to test exception behavior
  double xx = (3*start+end)/4;
  vv[0] = xx; 
  if ( s(xx) != s(vv) ) {
    std::cout << message << ":\n    s(xx) = " << s(xx) << " s(vv) =  "
	      << s(vv) << "\n";
    ok = false;
  }
  double h4 = h*h*h*h;
  double worst = 0;
  double x0;
  double x = start;
  const int intervals = 4357;
//  const int intervals = 100;
  double spacing = (end-start)/intervals;
  while (x < end) {
    double v_s = s(x);
    double v_f = f(x);
    double err = fabs(v_s - v_f);
    double scaled_err = err / h4; 
      			#ifdef VERBOSE_checkSpline
			std::cout << message << "  " << x << "  "
				  << scaled_err << "\n";
			#endif
    if ( err > worst ) {
      x0 = x;
      worst = err;
      			#ifdef TRACE_checkSpline
			std::cout << message << "  " << x << "  "
				  << scaled_err << "\n";
			#endif
      if ( scaled_err > 20 ) {
	#ifndef VERBOSE_checkSpline
	#ifndef TRACE_checkSpline
        std::cout << message << ":\n    x = " << x << " error = "
		  << err << " which scales to " << scaled_err << "\n";
	#endif
	#endif
	ok = false;
      }
    }
    x += spacing;
  }
      			#ifdef BRIEF_checkSpline
			std::cout << message << ": worst =  " << worst/h4 <<"\n";
			#endif
  return ok;
}


double sin2x   ( double x ) { return std::sin(2*x); }
double cos2pix ( double x ) { return std::cos(2*pi*x); }
double sin2pix ( double x ) { return std::sin(2*pi*x); }
double sin2xv  (std::vector<double> const & v) { return std::sin(2*v[0]); }
double sin2pixv  (std::vector<double> const & v) { return std::sin(2*pi*v[0]); }

class SinKx {
  public:
  SinKx(double kk) : k(kk) {}
  double operator()(double x) const { return std::sin(k*x); }
  double k;
};

class CosKx {
  public:
  CosKx(double kk) : k(kk) {}
  double operator()(double x) const { return std::cos(k*x); }
  double k;
};

double sin2pixValuesTable (int i) {
    double start = -1;
    double end   =  2;
    int gpoints = 101;
    double x = start + i * (end-start)/(gpoints-1);
    return (std::sin(2*pi*x));
}

int cos2pixTableSize = 101;

double cos2pixValuesTable (int i) {
    double start = -1;
    double end   =  2;
    int gpoints = cos2pixTableSize;
    double x = start + i * (end-start)/(gpoints-1);
    return (std::cos(2*pi*x));
}

double sin2pixValuesTableV (std::vector<int> const & v) {
    double start = -1;
    double end   =  2;
    int gpoints = cos2pixTableSize;
    double x = start + v[0] * (end-start)/(gpoints-1);
    return (std::sin(2*pi*x));
}

bool f_ctor () 
{ int gpoints = 2000;
  double start = -pi; 
  double end   = 2*pi;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  Spline<1> sin2xSpline ( sin2x, g );
  // sin(2x) is very well described by natural spline at -pi and pi
  // since the second derivatives at the endpoints are in fact zero.
  return checkSpline ( sin2xSpline, sin2x, start, end, 
  		      2 * (start-end)/(2*pi*(gpoints-1)),
		      "sin2x as a function" );
}

bool fewer_points ()
{ int gpoints = 200;					        
  double start = -pi;					        
  double end   = 2*pi;  				        
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);    
  Spline<1> sin2xSpline ( sin2x, g );			        
  return checkSpline ( sin2xSpline, sin2x, start, end,          
  		      2 * (end-start)/(2*pi*(gpoints-1)),      
  		      "sin2x as a function (200 points)" );    
}

bool outside_region ()
{ int gpoints = 2000;
  double start = -pi; 
  double end   = 2*pi;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  Spline<1> sin2xSpline ( sin2x, g );
  return checkSpline ( sin2xSpline, sin2x, start-(end-start)/gpoints, end, 
  		      2 * (end-start)/(2*pi*(gpoints-1)),
		      "sin2x with tiny outside region" );
}

bool ftor_irregular_grid () 
{ double gg[20] = { 0,  1,  3,    6,      7,  8.5, 12, 16, 17, 20,
                   21, 26, 26.5, 27.5,   28, 30,   33, 36, 39, 40 };
  Grid<1> g ( gg, gg+20 );
  int gpoints = 40;
  double start = 0; 
  double end   = 40;
  SinKx s ( pi/20 );
  Spline<1> sinKxSpline ( s, g );
  return checkSpline ( sinKxSpline, s, start, end, 
  		      5.0/40,
		      "spline on irregular grid from functor SinKx" );
}

bool natural_table ()
{ int gpoints = 101;
  double start = -1; 
  double end   = 2;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  Spline<1> sin2pixSpline ( sin2pixValuesTable , g );
  // The point is that for sin (2*pi*x) in this interval, natural boundary
  // conditions give precisely the right second derivatives at boundaries
  return checkSpline ( sin2pixSpline, sin2pix, start, end, 
  		     (end-start)/(gpoints-1),
		      "sin2pix as a table" );
}

bool interior_table ()
{ int gpoints = cos2pixTableSize;
    double start = -1; 
    double end   = 2;
    Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
    std::vector<double> cos2pixValues;
    for (int i=0; i< gpoints; ++i) {
      double x = start + i * (end-start)/(gpoints-1);
      cos2pixValues.push_back (std::cos(2*pi*x));
    }
    Spline<1> cos2pixSpline ( cos2pixValuesTable, g );
    int guardIntervals = 4;
    // By using "natural" boundary conditions (which are not accurate), 
    // we don't get excellent accuracy until four steps from the boundary.
    return checkSpline ( cos2pixSpline, cos2pix, 
    			start+ guardIntervals*(end-start)/(gpoints-1), 
			end  - guardIntervals*(end-start)/(gpoints-1), 
  		       (end-start)/(gpoints-1),
		        "cos2pix as a table" );
}

#ifdef TESTING_EMBEDDED_SPLINES
bool embedded_f ()
{ int gpoints = 101;
  double start = -1; 
  double end   = 2;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  CosKx c (2*pi);
  Spline<1> cos2pixSpline ( c, g, cos2pix );
  int guardIntervals = 1;
  // By using the derivatives at the boundaries, we get excellent
  // accuracy as close as one step from the boundary.
  return checkSpline ( cos2pixSpline, cos2pix, 
        	      start+ guardIntervals*(end-start)/(gpoints-1), 
		      end  - guardIntervals*(end-start)/(gpoints-1), 
  		     (end-start)/(gpoints-1),
		      "cos2pix as an embedded function" );
}
			  
bool embedded_ftor ()
{ int gpoints = 101;
  double start = -pi; 
  double end   = 2*pi;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  SinKx s (2);
  Spline<1> spline ( s, g, sin2x );
  int guardIntervals = -20;
  // The embedding should yield correct results outside the range of spline
  return checkSpline ( spline, sin2x, 
        	      start+ guardIntervals*(end-start)/(gpoints-1), 
		      end  - guardIntervals*(end-start)/(gpoints-1), 
  		     (end-start)/(gpoints-1),
		      "sin2x embedded in a ftor" );
}
			  
bool embedded_interior_table ()
{ int gpoints = cos2pixTableSize;
    double start = -1; 
    double end   = 2;
    Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
    std::vector<double> cos2pixValues;
    for (int i=0; i< gpoints; ++i) {
      double x = start + i * (end-start)/(gpoints-1);
      cos2pixValues.push_back (std::cos(2*pi*x));
    }
    CosKx c (2*pi);
    Spline<1> cos2pixSpline ( cos2pixValuesTable, g, c );
    int guardIntervals = 4;
    // By using "natural" boundary conditions (which are not accurate), 
    // we don't get excellent accuracy until four steps from the boundary.
    return checkSpline ( cos2pixSpline, cos2pix, 
    			start+ guardIntervals*(end-start)/(gpoints-1), 
			end  - guardIntervals*(end-start)/(gpoints-1), 
  		       (end-start)/(gpoints-1),
		        "cos2pix as an embedded interior table" );
}
#endif

bool generic_f ()
{ int gpoints = 200;					        
  double start = -pi;					        
  double end   = 2*pi;  				        
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);    
  Spline<1> sin2xSpline ( sin2xv, g );			        
  return checkSpline ( sin2xSpline, sin2x, start, end,          
  		      2 * (end-start)/(2*pi*(gpoints-1)),      
  		      "sin2x as a function of vector (200 points)" );    
}

bool generic_table ()
{ int gpoints = 101;
  double start = -1; 
  double end   = 2;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  Spline<1> sin2pixSpline ( sin2pixValuesTableV , g );
  return checkSpline ( sin2pixSpline, sin2pix, start, end, 
  		     (end-start)/(gpoints-1),
		      "sin2pix as a table based on vector" );
}

#ifdef TESTING_EMBEDDED_SPLINES
bool embedded_generic_table ()
{ int gpoints = 101;
  double start = -1; 
  double end   = 2;
  Grid<1> g (start, end + (end-start)/(gpoints-1), gpoints);  
  Spline<1> sin2pixSpline ( sin2pixValuesTableV , g, sin2pixv );
  return checkSpline ( sin2pixSpline, sin2pix, start, end, 
  		     (end-start)/(gpoints-1),
		      "sin2pix as an embedded interior table based on vector");
}
#endif

int main() {
  bool ok = true;

  ok &= f_ctor();
  ok &= fewer_points();
  ok &= outside_region();
  ok &= ftor_irregular_grid();
  ok &= natural_table();    
  ok &= interior_table();
  ok &= generic_f();
  ok &= generic_table();
#ifdef TESTING_EMBEDDED_SPLINES
  ok &= embedded_f();
  ok &= embedded_ftor();
  ok &= embedded_interior_table();
  ok &= embedded_generic_table();
#endif       
  return ok ? 0:-1;
}
